<?php
use Softneta\MedDream\Core\ActionFilter;
use Softneta\MedDream\Core\Backend;
use Softneta\MedDream\Core\HttpUtils;
use Softneta\MedDream\Core\ReportManager;
use Softneta\MedDream\Core\Logging;

define('PATH_TO_ROOT', __DIR__ . DIRECTORY_SEPARATOR);

require_once(PATH_TO_ROOT . 'autoload.php');

session_start();

$backend = new Backend(array('Auth', 'Report', 'Structure'));
if (!$backend->authDB->isAuthenticated())
    HttpUtils::error('Not Authenticated', 401);

$cmd = HttpUtils::getValueStrict($_REQUEST, 'cmd');

$reports = new ReportManager($backend->pacsReport, $backend->pacsStructure, $backend->attachmentUploadDir,
    $backend->authDB->getAuthUser());

$filter = new ActionFilter($backend->log);

function requireUploadPermission() {
    global $backend;
    if (!$backend->pacsAuth->hasPrivilege('upload'))
        HttpUtils::error('No permissions to upload', 403);
}

function getStudyMetadataAndVerifyAccess($studyUID) {
    global $reports;
    global $filter;

    $meta = $reports->getStudyMetadata($studyUID);
    $study = array();
    $study['error'] = isset($meta['error']) ? $meta['error'] : '';
    if (isset($meta['patientId'])) {
        $study['patientid'] =  $meta['patientId'];
    }

    /* will verify access after obtaining the data: in case of a PATIENT action, the only way
       is to check whether data contains the same Patient ID
     */
    if ($filter->getActionsInto($actions) &&
        $filter->actionsDoExist($actions) &&
        !$filter->validateAtStudyLevel('study', $studyUID, $actions, $study))
    {
        HttpUtils::error($study['error'], 403);
    }

    return $meta;
}

switch ($cmd) {
    case 'getTemplates':
        HttpUtils::returnJSON($reports->getTemplates());
        break;
    case 'getTemplate':
        header('Content-Type: text/plain');
        $id = HttpUtils::getValueStrict($_REQUEST, 'id');
        echo $reports->getTemplate($id);
        break;
    case 'saveTemplate':
        requireUploadPermission();
        HttpUtils::returnJSON(array(
            'id' => $reports->saveTemplate(
                HttpUtils::getValue($_REQUEST, 'id', -1),
                HttpUtils::getValueStrict($_REQUEST, 'group'),
                HttpUtils::getValueStrict($_REQUEST, 'name'),
                HttpUtils::getValueHtmlStrict($_REQUEST, 'body')
            )
        ));
        break;
    case 'removeTemplate':
        requireUploadPermission();
        $id = HttpUtils::getValueStrict($_REQUEST, 'id');
        $reports->removeTemplate($id);
        echo '{}';
        break;
    case 'getStudyMeta':
        $uid = HttpUtils::getValueStrict($_REQUEST, 'uid');
        $meta = getStudyMetadataAndVerifyAccess($uid);
        HttpUtils::returnJSON($meta);
        break;
    case 'getStudyReport':
        $uid = HttpUtils::getValueStrict($_REQUEST, 'uid');
        getStudyMetadataAndVerifyAccess($uid);
        $result = $reports->getStudyReport($uid);
        HttpUtils::returnJSON($result === false ? array() : $result);
        break;
    case 'getStudyReportContent':
        $uid = HttpUtils::getValueStrict($_REQUEST, 'uid');
        getStudyMetadataAndVerifyAccess($uid);
        $result = $reports->getStudyReportHTML($uid);
        if ($result === false)
            HttpUtils::error('Report Not Found', 404);
        echo $result;
    case 'saveStudyReport':
        requireUploadPermission();
        $uid = HttpUtils::getValueStrict($_REQUEST, 'uid');
        $content = HttpUtils::getValueHtmlStrict($_REQUEST, 'content');
        getStudyMetadataAndVerifyAccess($uid);
        HttpUtils::returnJSON($reports->saveStudyReport($uid, $content));
        break;
    case 'getAttachments':
        $uid = HttpUtils::getValueStrict($_REQUEST, 'uid');
        getStudyMetadataAndVerifyAccess($uid);
        HttpUtils::returnJSON($reports->getAttachments($uid, HttpUtils::getParam('id')));
        break;
    case 'downloadAttachment':
        $id = HttpUtils::getValueStrict($_REQUEST, 'id');
        $uid = HttpUtils::getValueStrict($_REQUEST, 'uid');
        getStudyMetadataAndVerifyAccess($uid);
        $reports->downloadAttachment($uid, $id);
        break;
    case 'removeAttachment':
        requireUploadPermission();
        $uid = HttpUtils::getValueStrict($_REQUEST, 'uid');
        $rid = HttpUtils::getValueStrict($_REQUEST, 'rid');
        $id = HttpUtils::getValueStrict($_REQUEST, 'id');
        getStudyMetadataAndVerifyAccess($uid);
        $reports->removeAttachment($uid, $rid, $id);
        echo '{}';
        break;
    case 'addAttachment':
        requireUploadPermission();
        if (empty($_FILES) || !isset($_FILES['attachment']))
            HttpUtils::error('File Not Specified');
        $uid = HttpUtils::getValueStrict($_REQUEST, 'uid');
        $id = HttpUtils::getValueStrict($_REQUEST, 'id');
        getStudyMetadataAndVerifyAccess($uid);
        HttpUtils::returnJSON($reports->addAttachment($uid, $id, $_FILES['attachment']));
        break;
    default:
        HttpUtils::error('invalid cmd');
}
